"""
Mamba interpreter ver 0.2
"""

import re

def process_line(line,r=0):
    # recursion dpeth

    # transaction definitions
    for match in re.finditer(r'\<\~\~\>\s+[\(\)A-Za-z0-9,\@\s\_]+',line):
        start, end = match.span()
        mstr = match.group(0)

        try:
            line = line.replace("<~~>", '').replace(" ", '')
        except:
            pass

    # replace agent references
    for match in re.finditer(r'\@[A-Za-z0-9_]+',line):
        start, end = match.span()
        mstr = match.group(0)
        line = line.replace(mstr,"self.%s" % mstr[1:]) # .lower()

    # power
    line = line.replace("^", "**")

    # time reference
    line = line.replace("$.TIME", "Clock().get_time()")

    # end
    line = line.replace("$[END]","\n")

    # self.
    line = line.replace("$.", "self.")

    # random number
    line = line.replace("$URAND", "np.random.rand()")


    # balance sheet stuff
    code = {
    "BALANCE?" : "self.balance_sheet.get_balance",
    "INCOME?": "self.income_statement.get_entry",
    "ASSETS": "BalanceEntry.ASSETS",
    "EQUITY": "BalanceEntry.EQUITY",
    "LIABILITIES": "BalanceEntry.LIABILITIES",
    "REVENUES": "ICSEntry.REVENUES",
    "GAINS":"ICSEntry.GAINS",
    "EXPENSES": "ICSEntry.EXPENSES",
    "LOSSES": "ICSEntry.LOSSES",
    "NONTAX_PROFITS": "ICSEntry.NONTAX_PROFITS",
    "INTEREST": "ICSEntry.INTEREST",
    "TAXES": "ICSEntry.TAXES",
    "NOI": "ICSEntry.NOI"
    }
    for k,v in code.items():
        line = line.replace(k,v)

    return (2+r)*"\t" + line + "\n"



def convert_code(codelines):

    mydict = {} # maps lines to code
    mycode_idx = 0

    import_code = """
from sfctools import World, Agent, Clock, Settings, BalanceEntry, ICSEntry
import numpy as np

from .transactions import *
"""

    knows_code = """
\t\tdef link(self):\n
"""

    mycode = ""

    r = 0 # recursion depth for nested statements

    line_idx = 0

    commend_mode = False # current line is a comment?

    for line in codelines:
        line = line.rstrip()
        if line.startswith("$[AGENT]"):  # agent definition starts

            try:
                mydef = line.split("$[AGENT]")[1].strip()
                #name = mydef[0].capitalize() + mydef[1:].split(".")[0]
                name = mydef[0] + mydef[1:].split(".")[0]

                my_args = mydef[1:].split(" ")

                cls_name = "Agent"

                if len(my_args) > 1:
                    cls_name = my_args[1]
                    cls_name = cls_name.replace("(","").replace(")","")
                    #name = mydef[0].capitalize() + my_args[0]
                    name = mydef[0] + my_args[0]

                mycode += """
class %s(%s):

\t\tdef __init__(self,*args):
\t\t\tsuper().__init__()
""" % (name,cls_name)
            except:

                return "ERROR please specify name after $[AGENT]",mydict

        
        elif line.startswith("$[CLASS]"):  # agent definition starts

            try:
                mydef = line.split("$[CLASS]")[1].strip()
                #name = mydef[0].capitalize() + mydef[1:].split(".")[0]
                name = mydef[0] + mydef[1:].split(".")[0]

                my_args = mydef[1:].split(" ")

                cls_name = ""

                if len(my_args) > 1:
                    cls_name = my_args[1]
                    cls_name = cls_name.replace("(","").replace(")","")
                    #name = mydef[0].capitalize() + my_args[0]
                    name = mydef[0] + my_args[0]

                mycode += """
class %s(%s):

\t\tdef __init__(self,*args):
\t\t\tsuper().__init__()
""" % (name,cls_name)
            except:

                return "ERROR please specify name after $[AGENT]",mydict

        elif line.startswith("$[MARKET]"):  # market definition starts

            try:
                mydef = line.split("$[MARKET]")[1].strip()
                #name = mydef[0].capitalize() + mydef[1:].split(".")[0]
                name = mydef[0] + mydef[1:].split(".")[0]

                my_args = mydef[1:].split(" ")

                cls_name = "MarketMatching"
                import_code += "from sfctools import MarketMatching\n"

                mycode += """
class %s(%s):

\t\tdef __init__(self):
\t\t\tsuper().__init__()
""" % (name,cls_name)
            except:

                return "ERROR please specify name after $[AGENT]",mydict


        elif line.startswith("+[PARAM]"):

            try:
                params = line.split("+[PARAM]")[1].strip()
                params = [p.strip() for p in params.split(",")]

                for p in params:
                    mycode += "\t\t\tself.%s = Settings().get_hyperparameter('%s')\n" % (p,p)
                mycode += "\n"

            except:
                pass

        elif line.startswith("+[KNOWS]"):

            try:
                knowns = line.split("+[KNOWS]")[1].strip()
                knowns = [k.strip() for k in knowns.split(",")]

                for k in knowns:
                    mycode += "\t\t\tself.%s = None\n" % (k)
                    # knows_code += "\t\t\tself.%s = World().get_agents_of_type('%s')\n" % (k.lower(),k.capitalize())
                    knows_code += "\t\t\tself.%s = World().get_agents_of_type('%s')\n" % (k,k)

                mycode += "\n" + knows_code + "\n"

            except:
                pass

        elif line.startswith("+[IMPORT]"):

            try:
                statement = line.split("+[IMPORT]")[1].strip()
                import_code += statement + "\n"

            except:
                pass

        elif line.startswith("+[IMPORT]"):

            try:
                statement = line.split("+[IMPORT]")[1].strip()
                import_code += statement + "\n"

            except:
                pass

        elif line.startswith("+[ENDFUN]") or line.strip().startswith("+[ENDACCOUNTING]"):
            r -= 1

        elif line.startswith("+[FUN]"):
            r += 1

            try:
                mydef = line.split("+[FUN]")[1]

                my_args = mydef.split(" ")
                fun_name = my_args[1]

                if len(my_args) > 2:

                    my_params = [s.replace("(","").replace(")","").strip() for s in my_args[2].split(",")]
                    mycode += "\t\tdef %s(self,%s):\n" % (fun_name, ",".join(my_params))

                else:

                    mycode += "\t\tdef %s(self):\n" % (fun_name)


            except:
                pass

        elif line.strip().startswith("+[ACCOUNTING]"):
            r += 1

            try:

                # beginning of
                mycode += "\t" + "\t"*(r+line.count("\t")) + "with self.balance_sheet.modify:\n"

            except:
                pass

        elif line.strip().startswith("<>"):

            try:
                s = line.split("<>")[1].strip()
                mycode += process_line("\t"*(r+line.count("\t")) +"self.balance_sheet.change_item(%s)"  % s)

            except:
                pass

        elif line.strip().startswith("<.>"):

            try:
                s = line.split("<.>")[1].strip()
                mycode += process_line("\t"*(r+line.count("\t")) +"self.balance_sheet.change_item(%s,suppress_stock=True)"  % s)

            except:
                pass

        elif line.startswith("+[INIT]"):
            r += 1

        elif line.startswith("+[ENDINIT]"):
            r -= 1


        else:
            mycode += process_line(line,r)


        mycode_idx = len(import_code.split("\n")) + len(mycode.split("\n"))
        mydict[line_idx] = mycode_idx
        line_idx += 1


        if line.startswith("$[END]"):
            all_code = import_code + "\n" + mycode
            all_code = all_code.encode('utf-8') # replace non utf-8
            all_code = str(all_code.decode("utf-8", "replace"))

            return all_code,mydict

    return "ERROR missing $[END]",mydict



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'prod' # options are: dev, test, prod


import yaml
# import dill
from ..core.singleton import Singleton
from ..bottomup.stock_manager import StockManager
from ..core.custom_warnings import warning_on_one_line

import dataclasses
from typing import Sequence
import pandas as pd
from collections import deque
import warnings
from typing import Any


class Settings(Singleton):
    """
    This is a settings class that serves as a central data structure to store hyperparameters, commodity names, depreciation rates etc.
    It also includes a section with metadata on author etc...

    """

    def __init__(self):
        """
        constructor for settings
        """
        super().__init__()  # <- has no effect (?)

        if hasattr(self, "initialized"): # <- TODO replace this with 'if self.do_init' from Singleton.
            return
        else:
            self.initialized = False # <- Tkept this because it works

            self.metainfo = None # meta informatino about project, author, date,...
            self.config_data = None  # this stores all information about allowed asset types
            self.df = None  # this stores the same data in dataFrame format (for convenience)
            self.hyper_params = None  # this stores the hyper parameters
            self.__hyper_params_dict = {}  # dict for faster access later
            self.depr_dict = {}  # asset and depreciation rate

            self.allowed_commodities = []
            self.unit_dict = {}  # asset and unit

            warnings.formatwarning = warning_on_one_line

    def __repr__(self):
        return "<Settings with %i entries>" % len(self.__hyper_params_dict)

    def __str__(self):
        return self.get_hyperparams_info().to_string()

    def reset(self):
        """
        reset settings
        """

        self.config_data = None
        self.metainfo = None
        self.df = None
        self.hyper_params = None
        self.__hyper_params_dict = {}  # dict for faster access later
        self.allowed_commodities = None
        self.depr_dict = {}

        warnings.warn("Settings have been reset!")

    def read_from_yaml(self,data,isfile=True):
        # warnings.warn("Settings().read_from_yaml is depreciated. use 'read' instead")

        return self.read(data,isfile)

    def read(self, data, isfile=True):
        """
        read the settings from a yaml file.
        The settings of a simulation can be stored in form of a yaml file. The first block metainfo stores the author, date and other meta data. The params block stores information on the assets allowed in balance sheets and their respective depreciation rates (0 means no depreciation, 1 means that the good is fully perishable). hyperparams contains parameters of agents stored centrally (for better overview). It also helps to avoid hardcoding of parameters directly in the code of the agents.

        :param data: either str, i.e. a string containing the data, or path/str, i.e. file name to read from
        :param isfile: boolean switch. If False, a string has to be passed.

        ** Example **

        "filename.yml" File:

        .. code-block:: yaml

            metainfo :
                author: Hans Dampf
                date: Agust 1964
                info: some more info

            params:
                - name: Cash
                  depreciation: 0.0
                  price: p
                  value: 1.0
                  unit: Euro
                  description: my description

                - name: Energy
                  ...

            hyperparams:
                - name: epsilon
                  value: 0.05
                  description: my description

                - name:
                  ...

        Python code:

        .. code-block:: python

            from sfctools import Settings
            my_settings = Settings().read_from_yaml("filename.yml")


        """

        if isfile:
            file = open(data)
            self.config_data = yaml.load(file, Loader=yaml.SafeLoader)
        else:
            self.config_data = yaml.safe_load(data)

        if self.config_data is None:
            raise RuntimeError("Something went wrong when reading the settings file. Data was None")

        if "params" not in self.config_data:
            self.config_data["params"] = [] # empty list if not available

        self.allowed_commodities = self.config_data["params"]

        for entry in self.allowed_commodities:
            self.depr_dict[entry["name"].strip()] = float(entry["depreciation"])

        self.hyper_params = self.config_data["hyperparams"]
        for entry in self.hyper_params:
            try:
                self.__hyper_params_dict[entry["name"].strip()] = float(entry["value"])
            except:
                self.__hyper_params_dict[entry["name"].strip()] = entry["value"]

        self.metainfo = self.config_data["metainfo"]

        # 3. feed stock manager with data
        self.setup_stock_manager()

        # 4. load world data
        # world = None
        # try:
        #    worldfile = self.config_data["worlddata"]
        #    with open(worldfile, "rb") as file:
        #        world = dill.load(file)
        # except Exception as e:
        #    print(str(e))
        #    # TODO correct error / exception handling
        # return world

        self.initialized = True
        return self

    def setup_stock_manager(self):
        """
        set up the stock manager from the data stored in the current settings. For more information on the stock manager, see StockManager class.
        This should not be called manually by the user. It is automatically called within 'read_from_yaml()'.
        """

        stock_mgr = StockManager()  # get instance of the stock manager

        # iterate through entries
        for k,entry in enumerate(self.config_data["params"]):

            try:

                # prepare the data
                name = entry["name"]

                if "price" in entry and "value" in entry and "unit" in entry:  # insert price, value and untit # TODO < how to resolve confusion between price (name of the price index) and value (current price value)?
                    price_name = entry["price"]
                    price_value = entry["value"]
                    self.unit_dict[name] = entry["unit"]

                    # write data to stock manager
                    stock_mgr.register(name, price_name, price_value)

                else:
                    if "name" in entry:
                        warnings.warn("I could not register %s at stock manager. 'price', 'value' and/or 'unit' key is missing." % entry["name"])
                    else:
                        warnings.warn("I could not register item %i at stock manager. 'name', 'price', 'value' and/or 'unit' key is missing." % k)

            except Exception as e:

                if "name" in entry:
                    warnings.warn("I could not register %s at stock manager. 'price','value' and/or 'unit' key is missing." % entry["name"])
                else:
                    warnings.warn("I could not register item %i at stock manager. Name missing" % k)


    def write(self,fname):
        """
        Writes the settings to a filename as yaml file
        """

        # mydata = {"metainfo": self.metainfo, "hyperparams": self.hyper_params}

        with open(fname, 'w') as outfile:
            yaml.dump(self.config_data, outfile, default_flow_style=False)


    def get_depr_info(self):
        """
        EXPERIMENTAL return info about assets and their depreciation rates

        :return: df, pandas dataframe with information about allowed assets

        Example

        .. code-block:: python

            from sfctools import Settings
            dfinfo = Settings().get_info()
            print(dfinfo)

        """
        df = pd.DataFrame(self.config_data["params"])

        df.rename(columns={
            "name": "Name",
            "depreciation": "Depreciation",
            "unit": "Unit",
            "value": "Initial Stock Value",
            "price": "Price Index",
            "description": "Description"
        }, inplace=True)

        df.set_index("Name", inplace=True)

        self.df = df  # for convenience save this here as 'duplicate' pandas format

        return df


    def get_info(self):
        # short name for HyperParams info
        return self.get_hyperparams_info()

    def get_hyperparams_info(self):
        """
        return info about hyperparameters

        :return: df, pandas dataframe with information about allowed assets

        Example

        .. code-block:: python

            from sfctools import Settings
            dfinfo = Settings().get_hyperparams_info()
            print(dfinfo)

        """

        if len(self.hyper_params) == 0:
            raise TypeError("No hyperparams found.")

        df = pd.DataFrame(self.hyper_params)
        df.rename(columns={"name": "Name", "value": "Value", "description": "Description"}, inplace=True)
        df.set_index("Name", inplace=True)

        return df

    def __getitem__(self,key):
        return self.get_hyperparameter(key)

    def __setitem__(self,key,value):
        warnings.warn("Changed settings parameter %s" %key)
        self.set_hyperparameter(key,value)

    def get_hyperparameter(self, name):
        """
        get the value of a hyperparameter by its name.

        :param name: the name of the hyper-parameter to filter for
        """

        if str(name) in self.__hyper_params_dict:
            return self.__hyper_params_dict[str(name)]  # faster access in O(1)
        else:

            err = "Could not find parameter %s in settings yaml. Please check again!" % (str(name))

            if not self.initialized:
                err += " Settings have not yet been initialized. Pleas use Settings().read(...) to load the simulation settings."

            raise KeyError(err)

    def set_hyperparameter(self,name,value):
        """
        get the value of a hyperparameter by its name.

        :param name: str, name of the parameter
        :param value: new value (any type but ideally same as before)
        """
        name = str(name)

        if name in self.__hyper_params_dict:
            # warnings.warn("Resetting hyperparameter value of %s" % name) < warning already given  in 'changed settings' warning

            self.__hyper_params_dict[str(name)] = value

            for i in self.config_data["hyperparams"]:
                if i["name"] == name:
                    if isinstance(value,float) or isinstance(value,int):
                        i["value"] = float(value) # ints will be converted to floats
                    else: # anything else
                        i["value"] = value

                    break
        else:

            err = "Could not find parameter %s in settings yaml. Please check again!" % (str(name))

            if not self.initialized:
                err += " Settings have not yet been initialized. Pleas use Settings().read(...) to load the simulation settings."

            raise KeyError(err)



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'dev' # options are: dev, test, prod

from sfctools import FlowMatrix as FlowLogger
import numpy as np
from sfctools import BalanceEntry,Accounts
from sfctools import ICSEntry
CA = Accounts.CA
KA = Accounts.KA


def con(household, production, q):
    #
    assert not np.isnan(q)
    assert q < +np.inf
    assert q > -np.inf
    assert q >= 0, print('have to pass positive quantity: unidirectional transaction')
    household.balance_sheet.disengage()
    production.balance_sheet.disengage()
    household.balance_sheet.change_item('Equity', BalanceEntry.EQUITY, -q, suppress_stock=False)
    household.balance_sheet.change_item('Cash', BalanceEntry.ASSETS, -q, suppress_stock=False)
    production.balance_sheet.change_item('Equity', BalanceEntry.EQUITY, +q, suppress_stock=False)
    production.balance_sheet.change_item('Cash', BalanceEntry.ASSETS, +q, suppress_stock=False)
    household.balance_sheet.engage()
    production.balance_sheet.engage()
    FlowLogger().log_flow((KA,KA), q, household, production, subject='Consumption')
    household.income_statement.new_entry(ICSEntry.EXPENSES,'con',q)
    production.income_statement.new_entry(ICSEntry.REVENUES,'con',q)


def gov(government, production, q):
    #government consumption
    assert not np.isnan(q)
    assert q < +np.inf
    assert q > -np.inf
    assert q >= 0, print('have to pass positive quantity: unidirectional transaction')
    government.balance_sheet.disengage()
    production.balance_sheet.disengage()
    government.balance_sheet.change_item('Equity', BalanceEntry.EQUITY, -q, suppress_stock=False)
    government.balance_sheet.change_item('Cash', BalanceEntry.ASSETS, -q, suppress_stock=False)
    production.balance_sheet.change_item('Equity', BalanceEntry.EQUITY, +q, suppress_stock=False)
    production.balance_sheet.change_item('Cash', BalanceEntry.ASSETS, +q, suppress_stock=False)
    government.balance_sheet.engage()
    production.balance_sheet.engage()
    FlowLogger().log_flow((KA,KA), q, government, production, subject='Gov. Expenditure')
    government.income_statement.new_entry(ICSEntry.EXPENSES,'gov',q)
    production.income_statement.new_entry(ICSEntry.NONTAX_PROFITS,'gov',q)


def wag(production, household, q):
    #
    assert not np.isnan(q)
    assert q < +np.inf
    assert q > -np.inf
    assert q >= 0, print('have to pass positive quantity: unidirectional transaction')
    production.balance_sheet.disengage()
    household.balance_sheet.disengage()
    production.balance_sheet.change_item('Equity', BalanceEntry.EQUITY, -q, suppress_stock=False)
    production.balance_sheet.change_item('Cash', BalanceEntry.ASSETS, -q, suppress_stock=False)
    household.balance_sheet.change_item('Equity', BalanceEntry.EQUITY, +q, suppress_stock=False)
    household.balance_sheet.change_item('Cash', BalanceEntry.ASSETS, +q, suppress_stock=False)
    production.balance_sheet.engage()
    household.balance_sheet.engage()
    FlowLogger().log_flow((KA,KA), q, production, household, subject='Wages')
    production.income_statement.new_entry(ICSEntry.EXPENSES,'wag',q)
    household.income_statement.new_entry(ICSEntry.REVENUES,'wag',q)


def tax(production, government, q):
    # taxes on firm profits
    assert not np.isnan(q)
    assert q < +np.inf
    assert q > -np.inf
    assert q >= 0, print('have to pass positive quantity: unidirectional transaction')
    production.balance_sheet.disengage()
    government.balance_sheet.disengage()
    production.balance_sheet.change_item('Equity', BalanceEntry.EQUITY, -q, suppress_stock=False)
    production.balance_sheet.change_item('Cash', BalanceEntry.ASSETS, -q, suppress_stock=False)
    government.balance_sheet.change_item('Equity', BalanceEntry.EQUITY, +q, suppress_stock=False)
    government.balance_sheet.change_item('Cash', BalanceEntry.ASSETS, +q, suppress_stock=False)
    production.balance_sheet.engage()
    government.balance_sheet.engage()
    FlowLogger().log_flow((KA,KA), q, production, government, subject='Taxes')
    production.income_statement.new_entry(ICSEntry.TAXES,'tax',q)
    government.income_statement.new_entry(ICSEntry.REVENUES,'tax',q)



from PyQt5 import QtWidgets, uic,QtGui
from PyQt5.QtCore import QUrl
from PyQt5.QtWidgets import QFileDialog,QMessageBox,QTableWidgetItem,QAbstractItemView,QListWidgetItem
import sys
import os
import yaml
import shutil
import time
import datetime

# from pandasmodel import PandasModel
from PyQt5.QtGui import QDesktopServices
from PyQt5.QtCore import QObject, pyqtSlot
import pickle as pkl
import re
import matplotlib.pyplot as plt
from PyQt5.QtCore import Qt
from PyQt5 import QtCore
from PyQt5.QtGui import QFont, QFontDatabase, QColor, QSyntaxHighlighter, QTextCharFormat
from PyQt5.QtCore import QObject, QThread, pyqtSignal

from PyQt5.QtWidgets import QShortcut
from PyQt5.QtGui import QKeySequence


class MainWatchDog(QObject):
    """
    This is a file watchdog. It frequently checks if external changes
    have been made to a file viewed in the code editor.
    """
    difference_detected = pyqtSignal()  # difference detected between current code and code stored in file

    def __init__(self,parent, file_ctrl):

        super().__init__()
        #self.file_name = os.path.join(file_name , agent_name + ".txt" ) # < source file sub_names

        self.parent_w = parent
        text = parent.textEdit.toPlainText()
        # self.finished = False
        self.file_ctrl = file_ctrl  # <-  set mutable type finished list

    def run(self):
        """Long-running task."""

        while True:

            time.sleep(1.0) # check every second

            if self.file_ctrl[0] is None:
                continue

            file_name = os.path.join(os.path.dirname(self.file_ctrl[0]) ,"python_code", "mainloop.py" ) # < source file sub_names

            # print("[Watchdog] check file", file_name)
            current_gui_text = self.parent_w.textEdit.toPlainText()
            current_file_text = None

            try:
                with open(file_name,"r") as file:
                    current_file_text = file.read()

                outdated = False
                file_time = os.path.getmtime(file_name)
                my_time = self.parent_w.mtime

                if MainLoopEditor.instance is None:
                    self.finished = True

                if file_time > my_time:
                    outdated = True

                if current_gui_text != current_file_text and outdated:
                    self.difference_detected.emit()

            except Exception as e:
                print(str(e))
                # pass



class Highlighter(QSyntaxHighlighter):
    def __init__(self,parent=None):
        super().__init__(parent)
        self._mapping = {}

    def add_mapping(self, pattern, pattern_format):
        self._mapping[pattern] = pattern_format

    def highlightBlock(self, text_block):
        for pattern, fmt in self._mapping.items():
            try:

                for match in re.finditer(pattern,text_block):
                    try:
                        start, end = match.span()
                        self.setFormat(start, end-start, fmt)
                    except:
                        pass

            except:
                pass

class MainLoopEditor(QtWidgets.QDialog):

    instance = None

    def __init__(self,parent,text=""):

        super(MainLoopEditor, self).__init__(parent) # Call the inherited classes __init__ method
        path = os.path.dirname(os.path.abspath(__file__))
        uic.loadUi(os.path.join(path,'simulation_edit.ui'), self) # Load the .ui file

        self.setFixedSize(self.size());

        self.backup_count = 0

        self.textEdit.textChanged.connect(self.text_changed)
        self.textEdit.cursorPositionChanged.connect(self.update_line_label)

        self.textEdit.installEventFilter(self)

        self.highlighter = Highlighter()

        self.setUpEditor()

        self.shortcut = QShortcut(QKeySequence("Ctrl+S"),self)
        self.shortcut.activated.connect(self.parent().save_and_build)

        self.shortcut2 = QShortcut(QKeySequence("Ctrl+Tab"),self)
        self.shortcut2.activated.connect(self.indent)

        self.shortcut2 = QShortcut(QKeySequence("Shift+Tab"),self)
        self.shortcut2.activated.connect(self.unindent)

        self.set_text(text)
        self.searchEdit.textChanged.connect(self.update_search)

        self.parent_widget = parent
        theme_manager = self.parent_widget.theme_manager
        self.setStyleSheet(theme_manager.get_background_style())
        self.setStyleSheet(theme_manager.get_stylesheet("main"))

        project_file = self.parent_widget.current_file
        print("Project File", project_file)

        self.__class__.instance = self
        self.thread = QThread()

        filepath = None # os.path.dirname(os.path.abspath(self.parent_widget.current_file))
        # python_path = os.path.join(filepath , "python_code")
        self.mtime = 0.0 # time.time() # os.path.getmtime(filepath) # last changes of project file

        self.file_ctrl = [filepath]
        self.worker = MainWatchDog(self, self.file_ctrl)
        self.worker.moveToThread(self.thread)
        self.worker.difference_detected.connect(self.process_differences)
        self.thread.started.connect(self.worker.run)

        self.thread.start()

    def start_watchdog(self,fname):
        # start a new watchdog on the file
        self.file_ctrl[0] = fname


    def set_text(self,text):
        text = text.replace("    ","\t")
        self.textEdit.setPlainText(text)

    def closeEvent(self, event):
        self.__class__.instance = None
        self.hide()
        # event.accept() # let the window close

    def process_differences(self):
        # process differences between the sfctl file and the external file changes
        self.warn_differences()
        self.set_text_from_external()

    def warn_differences(self):
        # self.main_widget.notify("Difference Detected!\n%s" % self.agent_name,title="External changes detected.")
        self.parent_widget.statusBar().showMessage("External changes in mainloop.py detected! Overwriting mainloop.py")

    def set_text_from_external(self):
        """
        set the text from the external file changes. Overwrite internal text
        """

        filepath = os.path.dirname(os.path.abspath(self.parent_widget.current_file))
        python_path = os.path.join(filepath , "python_code")

        new_text = None
        with open(os.path.join(python_path , "mainloop.py"),"r") as file:
            new_text = file.read()

        if new_text is None:
            return

        scroll_pos = self.textEdit.verticalScrollBar().value()

        cursor = self.textEdit.textCursor()
        old_position= cursor.position()

        self.textEdit.setPlainText(new_text)

        cursor = self.textEdit.textCursor()
        cursor.setPosition(old_position)

        self.textEdit.setTextCursor(cursor)
        self.textEdit.verticalScrollBar().setValue(scroll_pos)


    def update_search(self):
        self.setUpEditor()
        self.update()

    def eventFilter(self, obj, event):
        self.update()
        if event.type() == QtCore.QEvent.KeyPress and obj is self.textEdit:

            try:
                self.mtime = time.time()
            except Exception as e:
                print(str(e))

            if event.key() == QtCore.Qt.Key_Tab: # and self.text_box.hasFocus():
                print('Tab pressed')
                self.indent()
                return True
        return False

    def indent(self):
        # inert tab for whole selection

        scroll_pos = self.textEdit.verticalScrollBar().value()

        cursor = self.textEdit.textCursor()
        old_position= cursor.position()

        start = cursor.selectionStart()
        end = cursor.selectionEnd()

        text = self.textEdit.toPlainText()

        mystr = text[:start]
        my_lines = []

        for line in text[start:end].split("\n"):
            my_lines.append("\t" + line)

        new_str =  "\n".join(my_lines)
        new_end = start + len(new_str)
        mystr += new_str

        mystr += text[end:]

        self.textEdit.setPlainText(mystr)

        cursor = self.textEdit.textCursor()
        if new_end - start > 1:

            cursor.setPosition(start)
            cursor.setPosition(new_end, QtGui.QTextCursor.KeepAnchor)
        else:
            cursor.setPosition(new_end)

        self.textEdit.setTextCursor(cursor)
        self.textEdit.verticalScrollBar().setValue(scroll_pos)

    def unindent(self):
        # inert tab for whole selection

        scroll_pos = self.textEdit.verticalScrollBar().value()

        cursor = self.textEdit.textCursor()
        old_position= cursor.position()

        start = cursor.selectionStart()
        end = cursor.selectionEnd()

        text = self.textEdit.toPlainText()

        mystr = text[:start]
        my_lines = []

        for line in text[start:end].split("\n"):
            if line.startswith("\t"):
                my_lines.append(line[1:])
            else:
                my_lines.append(line)

        new_str =  "\n".join(my_lines)
        new_end = start + len(new_str)
        mystr += new_str

        mystr += text[end:]

        self.textEdit.setPlainText(mystr)

        cursor = self.textEdit.textCursor()

        if new_end - start > 1:

            cursor.setPosition(start)
            cursor.setPosition(new_end, QtGui.QTextCursor.KeepAnchor)
        else:
            cursor.setPosition(new_end)

        self.textEdit.setTextCursor(cursor)
        self.textEdit.verticalScrollBar().setValue(scroll_pos)

    def text_changed(self):
        self.send_data()
        self.update_line_label()
        self.update()

    def update_line_label(self):
        cursor = self.textEdit.textCursor()
        x = cursor.blockNumber() + 1
        y = cursor.columnNumber() + 1
        self.lineLabel.setText('{0:d}, {1:d}'.format(x,y))

    def send_data(self):
        self.parent().mainloop_str = self.textEdit.toPlainText()
        if self.backup_count == 10:
            self.parent().auto_backup()
            self.backup_count = 0
        self.backup_count += 1

    def paintEvent(self,event):
        qp = QtGui.QPainter(self)
        pen = QtGui.QPen(Qt.black ,1, Qt.SolidLine)
        brush = QtGui.QBrush(Qt.black)
        qp.setBrush(brush)
        qp.setPen(pen)

        pen = QtGui.QPen(QtGui.QColor("#ababab") ,1, Qt.SolidLine)
        qp.setPen(pen)
        brush = QtGui.QBrush(QtGui.QColor("#ababab"))
        qp.setBrush(brush)

        positions = []
        for block_number in range(self.textEdit.document().lineCount()):

            offset = self.textEdit.contentOffset()

            block= self.textEdit.document().findBlockByNumber(block_number)
            brect = self.textEdit.blockBoundingGeometry(block)
            # print("brect", brect)
            y = int(brect.y()+ offset.y()) + 20  #  53
            x = int(brect.x() + 6)
            w = brect.width()
            h = brect.height()

            if(y > 20 and y <= 820):
                try:
                    if y not in positions:
                        qp.drawText(x,y+ 4,"%04i" % int(block_number))
                        positions.append(y)
                except:
                    pass

    def setUpEditor(self):
        self.highlighter = Highlighter()
        # python stuff
        fmt = QTextCharFormat()
        fmt.setForeground(QtGui.QColor("#ff3838"))
        for pattern in ['def ','for ', 'if ', 'else:', "while ", ' in ', ' not ', 'elif ', "True", "False", "class "]:
            self.highlighter.add_mapping(pattern,fmt)

        # return statement
        fmt = QTextCharFormat()
        fmt.setForeground(QtGui.QColor("#dae0dc"))
        for pattern in ['return']:
            self.highlighter.add_mapping(pattern,fmt)

        #print
        fmt = QTextCharFormat()
        fmt.setForeground(QtGui.QColor("#3f88e8"))
        for pattern in ['print']:
            self.highlighter.add_mapping(pattern,fmt)

        # python stuff
        fmt = QTextCharFormat()
        fmt.setForeground(QtGui.QColor("#4bb53f"))
        for pattern in ['from ','import ', 'as ']:
            self.highlighter.add_mapping(pattern,fmt)

        # comments
        fmt = QTextCharFormat()
        fmt.setForeground(QtGui.QColor("#121212"))
        self.highlighter.add_mapping(r'\"{3}[A-Za-z0-9_\n\t\wx^^    ]+\"{3}',fmt)

        fmt = QTextCharFormat()
        # fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#6a6a6a"))
        pattern = r'\#\s.+'
        self.highlighter.add_mapping(pattern,fmt)

        searchtext = self.searchEdit.text()

        if searchtext != "":
            try:
                fmt = QTextCharFormat()
                # fmt.setFontItalic(True)
                fmt.setFontWeight(QFont.Bold)
                fmt.setForeground(QtGui.QColor("#fcf75b"))
                fmt.setBackground(QtGui.QColor("#000000"))
                self.highlighter.add_mapping(searchtext,fmt)
            except Exception as e:
                print(str(e))

        self.highlighter.setDocument(self.textEdit.document())



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'dev' # options are: dev, test, prod


from sfctools import SingletonAgent, Settings, World, BalanceEntry
from .transactions import gov as transactions_government


class Government(SingletonAgent):
    """
    The government agent consumes 
    """

    def __init__(self):
        """
        constructor for the government
        """
        super().__init__(alias="Government") # there is only one household
        
        self.G = Settings().get_hyperparameter("G")
        self.production = None 

    def link(self):
        """
        Links the government to the production agent
        """
        self.production = World().get_agents_of_type("Production")[0]

    def reset(self):
        """
        Resets the agent for the next simulation period
        """
        self.income_statement.reset()

    def update(self):
        """
        Updates the government
        """

        with self.balance_sheet.modify:

            self.balance_sheet.change_item("Cash",BalanceEntry.ASSETS,self.G)
            # print new money
            self.balance_sheet.change_item("Equity",BalanceEntry.EQUITY,self.G)
            # print new money
        
        # transfer government expenditure to production agent 
        transactions_government(self,self.production,self.G) 
    
    def file_bankruptcy(self,event):
        # print money if bankrupt 

        cb = -self.cash_balance

        with self.balance_sheet.modify:

            self.balance_sheet.change_item("Cash",BalanceEntry.ASSETS,cb) # print new money
            self.balance_sheet.change_item("Bonds",BalanceEntry.EQUITY,cb) # print new money
        




__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'dev' # options are: dev, test, prod


from sfctools import SingletonAgent, Settings, World
from sfctools import BalanceEntry

from .transactions import con as transaction_consumption 


class Household(SingletonAgent):
    """
    The household agent consumes 
    """

    def __init__(self):
        """
        Constructor for the household
        """
        super().__init__(alias="Household") # there is only one household

        self.C = 0.8  # initial consumption 
        self.DI = 0.0 # initial disposable income 

        self.alpha1 = Settings().get_hyperparameter("alpha1")
        self.alpha2 = Settings().get_hyperparameter("alpha2")

        self.production = None # <- placeholder for production agent 
    

    def link(self):
        """
        Links the household agent with the production agent 
        """
        self.production = World().get_agents_of_type("Production")[0]

        # equip household with some initial net worth 
        with self.balance_sheet.modify:
            self.balance_sheet.change_item("Cash",BalanceEntry.ASSETS, 25.0, suppress_stock=True)
            self.balance_sheet.change_item("Equity",BalanceEntry.EQUITY, 25.0,suppress_stock=True)

        # equip household with some initial net worth 
        with self.balance_sheet.modify:
            self.balance_sheet.change_item("Cash",BalanceEntry.ASSETS, 25.0, suppress_stock=True)
            self.balance_sheet.change_item("Bonds",BalanceEntry.LIABILITIES, 25.0,suppress_stock=True)

    def reset(self):
        """
        Resets the agent for the next simulation period
        """
        self.income_statement.reset()
    
    def update(self):
        """
        Updates the household agent
        """
        
        #print(self.balance_sheet.to_string())

        
        # retrieve wealth and income
        wealth = self.net_worth 
        income = self.income_statement.net_income

        self.DI = wealth+ income  # disposable income 

        # consumption 
        self.C = self.alpha1 * income + self.alpha2 * wealth

        # do the actual transaction
        transaction_consumption(self,self.production,self.C)



from sfctools.gui import Gui
from importlib.metadata import version
import sys
import os

if len(sys.argv) > 1 and sys.argv[1] == "attune":
    g = Gui()
    g.run()

elif len(sys.argv) > 1 and sys.argv[1] == "mamba-convert":
    # convert MAMBA source to Python and overwrite python_code folder

    source_folder = os.path.join(os.getcwd(),"mamba_code")
    dest_folder = os.path.join(os.getcwd(),"python_code")

    if not os.path.isdir(source_folder):
        print("ERROR: Could not detect source folder in %s" % source_folder)
    if not os.path.isdir(dest_folder):
        print("ERROR: COuld not detect destination folder in %s" % dest_folder)

    from sfctools.gui.attune.src.mamba_interpreter2 import convert_code

    # convert all files
    sub_names = next(os.walk(source_folder), (None, None, []))[2]

    k = 0
    for fname in sub_names:
        k+= 1
        try:

            all_code = None
            source_file = os.path.join(source_folder,fname)
            print("[%03i] %s" % (k,fname))
            with open(source_file,"r") as file:
                codelines = file.readlines()
                all_code,_ = convert_code(codelines)

            if all_code is not None:
                name = os.path.basename(fname)
                dest_file = os.path.join(dest_folder, name).lower()
                with open(dest_file,"w") as file:
                    file.write(all_code)

        except Exception as e:
            print("Error: Could not convert %s: %s" % (fname,str(e)))

    print("done.")

else:
    print("""
============= WELCOME TO SFCTOOLS ============================
ver %s
Main corresponding author: Thomas, thomas.baldauf@dlr.de
Institute of Networked Energy Systems (DLR-VE)
German Aerospace Center, 2020

Sfctools is a lightweight and easy-to-use Python framework
for stock-flow consistent agent-based macroeconomic (SFC-ABM) modeling.
It concentrates on agents in economics and helps you to construct agents,
helps you to manage and document your model parameters,
assures stock-flow consistency, and facilitates basic economic
data structures (such as the balance sheet).

Become part of the community! ;-)

See https://sfctools-framework.readthedocs.io/en/latest/ for the latest documentation
See https://gitlab.com/dlr-ve/esy/sfctools for the latest version on gitlab
==============================================================

Type 'python -m sfctools attune' to start the graphcial user interface 'attune'
Type 'python -m sfctools mamba-convert' to build a Python project from the MAMBA code folder
""" % version('sfctools'))



from PyQt5 import QtWidgets, uic,QtGui
from PyQt5.QtCore import QUrl
from PyQt5.QtWidgets import QFileDialog,QMessageBox,QTableWidgetItem,QAbstractItemView,QListWidgetItem
import sys
import os
import yaml
import shutil
# from pandasmodel import PandasModel
from PyQt5.QtGui import QDesktopServices
import pickle as pkl
import re
import matplotlib.pyplot as plt
from PyQt5.QtCore import Qt

from sfctools import Settings
from PyQt5.QtGui import QFont, QFontDatabase, QColor, QSyntaxHighlighter, QTextCharFormat
import pyperclip
from PyQt5.QtCore import QObject, pyqtSlot,pyqtSignal,QThread

from PyQt5.QtWidgets import QShortcut
from PyQt5.QtGui import QKeySequence

from .pandasmodel import PandasModel
import numpy as np
import time


class SettingsWatchDog(QObject):
    """
    This is a file watchdog. It frequently checks if external changes
    have been made to a file viewed in the settings (yaml) editor.
    """
    difference_detected = pyqtSignal()  # difference detected between current code and code stored in file

    def __init__(self, parent, file_ctrl):

        super().__init__()

        self.parent_w = parent
        text = parent.textEdit.toPlainText()
        self.file_ctrl = file_ctrl  # <-  set mutable type finished list

    def run(self):
        """Long-running task."""

        while True:

            time.sleep(1.0) # check every second

            if self.file_ctrl[0] is None:
                continue

            file_name = os.path.join(os.path.dirname(self.file_ctrl[0]) , "settings.yml" ) # < source file sub_names

            # print("[Watchdog] check file", file_name)
            current_gui_text = self.parent_w.textEdit.toPlainText()
            current_file_text = None

            try:
                with open(file_name,"r") as file:
                    current_file_text = file.read()

                outdated = False
                file_time = os.path.getmtime(file_name)
                my_time = self.parent_w.mtime

                if SettingsEditor.instance is None:
                    self.finished = True

                if file_time > my_time:
                    outdated = True

                if current_gui_text != current_file_text and outdated:
                    self.difference_detected.emit()
                    self.parent_w.mtime = time.time()

            except Exception as e:
                print(str(e))
                # pass



class Highlighter(QSyntaxHighlighter):
    def __init__(self,parent=None):
        super().__init__(parent)
        self._mapping = {}

    def add_mapping(self, pattern, pattern_format):
        self._mapping[pattern] = pattern_format

    def highlightBlock(self, text_block):

        for pattern, fmt in self._mapping.items():
            for match in re.finditer(pattern,text_block):
                start, end = match.span()
                self.setFormat(start, end-start, fmt)

class SettingsEditor(QtWidgets.QDialog):
    instance = None

    def __init__(self,parent=None,text="<no text found>"):

        super(SettingsEditor, self).__init__(parent) # Call the inherited classes __init__ method
        path = os.path.dirname(os.path.abspath(__file__))
        uic.loadUi(os.path.join(path,'settings_edit.ui'), self) # Load the .ui file

        self.backup_count = 0

        self.edit_count = 0

        self.parent_widget = parent

        # self.setFixedSize(self.size())
        self.fmt = self.textEdit.currentCharFormat()
        self.highlighter = Highlighter()
        self.setUpEditor()

        self.pushButton.pressed.connect(self.copy_data)

        self.shortcut = QShortcut(QKeySequence("Ctrl+S"),self)
        self.shortcut.activated.connect(self.rebuild_table)

        self.textEdit.setPlainText(text)
        self.textEdit.textChanged.connect(self.update_text)

        self.tableView.setColumnWidth(0,30)
        self.tableView.setColumnWidth(1,300)

        self.searchEdit.textChanged.connect(self.find_param)
        self.tableView.doubleClicked.connect(self.scroll_to)
        # self.okButton.pressed.connect(self.send_data)

        self.update_valid()
        self.rebuild_table()

        self.tableBtn.pressed.connect(self.rebuild_table)
        self.btnValid.pressed.connect(self.update_valid)

        self.__class__.instance = self
        self.thread = QThread()

        filepath = None
        self.mtime = 0 # time.time() # os.path.getmtime(filepath) # last changes of project file
        # ^ensure the time upon opening the yaml editor is lower than any other time

        self.file_ctrl = [filepath]
        self.worker = SettingsWatchDog(self, self.file_ctrl)
        self.worker.moveToThread(self.thread)
        self.worker.difference_detected.connect(self.process_differences)
        self.thread.started.connect(self.worker.run)

        self.thread.start()

        self.show()

    def process_differences(self):
        # process differences between the sfctl file and the external file changes
        self.warn_differences()
        self.set_text_from_external()

    def warn_differences(self):
        # self.main_widget.notify("Difference Detected!\n%s" % self.agent_name,title="External changes detected.")
        self.parent_widget.statusBar().showMessage("External changes in settings.yml detected! Overwriting settings.yml")

    def set_text_from_external(self):
        """
        set the text from the external file changes. Overwrite internal text
        """

        filepath = os.path.dirname(os.path.abspath(self.parent_widget.current_file))
        project_path = filepath # os.path.join(filepath , "python_code")

        new_text = None
        with open(os.path.join(project_path , "settings.yml"),"r") as file:
            new_text = file.read()

        if new_text is None:
            return

        scroll_pos = self.textEdit.verticalScrollBar().value()

        cursor = self.textEdit.textCursor()
        old_position= cursor.position()

        self.textEdit.setPlainText(new_text)

        cursor = self.textEdit.textCursor()
        cursor.setPosition(old_position)

        self.textEdit.setTextCursor(cursor)
        self.textEdit.verticalScrollBar().setValue(scroll_pos)


    def closeEvent(self, event):
        self.__class__.instance = None
        self.hide()
        # event.accept() # let the window close

    def start_watchdog(self,fname):
        # start a new watchdog on the file
        self.file_ctrl[0] = fname

    def update_text(self):
        # if there were more than 20 difs, check if settings are still valid automatically
        # cannot do this every time because it slows down the textedit a lot
        self.edit_count += 1
        if self.edit_count >= 1: # this caused some problems. try to do update_valid in separate thread every x seconds?
            self.edit_count = 0
            self.update_valid()

    def find_param(self):
        # find the parameter in the search field in the table

        df = Settings().get_hyperparams_info()
        my_index = list(df.index)
        search_word = self.searchEdit.text()

        if search_word in my_index:

            scroll_idx = my_index.index(search_word)
            try:
                self.tableView.selectRow(scroll_idx)
            except Exception as e:
                print(str(e))
            self.scroll_to()

    def rebuild_table(self):
        self.update_valid()
        # self.parent().save_and_build()

        #scroll_pos = self.textEdit.verticalScrollBar().value()
        try:
            model = PandasModel(Settings().get_hyperparams_info())
            self.tableView.setModel(model)
        except:
            self.parent().notify(msg="Something went wrong. Is the project empty?",title="Error")
        #self.textEdit.verticalScrollBar().setValue(scroll_pos)

        self.mtime = time.time()

    def scroll_to(self):
        # scroll to selected table item in te settings editor

        selected_rows = sorted(set(index.row() for index in
                          self.tableView.selectedIndexes()))

        if len(selected_rows) == 0:
            print("no selected rows")
            return

        selection = selected_rows[0]
        df = Settings().get_hyperparams_info()

        param_name = df.index[selection]
        print("PARAM_NAME",param_name)

        # find the parameter in the settings file left
        try:
            mytext = self.textEdit.toPlainText()
            for i,line in enumerate(mytext.split("\n")):
                if line.find("name: " + param_name) > -1:
                    minimum = self.textEdit.verticalScrollBar().minimum()
                    maximum = self.textEdit.verticalScrollBar().maximum()
                    self.textEdit.verticalScrollBar().setValue(int(np.clip(i,minimum,maximum)))
                    break

        except Exception as e:
            print(str(e))

    def copy_data(self):
        data = ""

        try:
            my_teststr = self.textEdit.toPlainText().replace("\t","    ")
            Settings().read(my_teststr,isfile=False)

            print(Settings().get_hyperparams_info())
            Settings().get_hyperparams_info().to_clipboard()

        except Exception as e:
            print("Error: " + str(e))


    def update_valid(self):
        # try to read this with yaml
        valid = True

        my_teststr = self.textEdit.toPlainText().replace("\t","    ")
        # yaml allows no tabs

        try:
            Settings().read(my_teststr,isfile=False)
        except:
            valid = False

        if valid:
            self.validLabel.setText("valid")
            self.validLabel.setStyleSheet("background-color: lightgreen; color: black;")
        else:
            self.validLabel.setText("invalid")
            self.validLabel.setStyleSheet("background-color: yellow; color: black;")

        self.update()

        self.send_data()



    def send_data(self):
        print("yaml editor send data")
        self.parent().settings_str = self.textEdit.toPlainText().replace("\t","    ")
        # self.close()
        if self.backup_count == 10:
            self.parent().auto_backup()
            self.backup_count = 0
        self.backup_count += 1

    def setUpEditor(self):


        for keyword in ["metainfo","params","hyperparams",
                         "depreciation", "price", "unit", "description"]:
            fmt = QTextCharFormat()
            fmt.setFontWeight(QFont.Bold)
            fmt.setForeground(QtGui.QColor("#f5f5f5"))
            pattern = r'%s\:' % keyword
            self.highlighter.add_mapping(pattern,fmt)

        for keyword in ["author", "date","info", "metainfo"]:
            fmt = QTextCharFormat()
            fmt.setFontWeight(QFont.Bold)
            fmt.setForeground(QtGui.QColor("#f5f5f5"))
            pattern = r'%s\:' % keyword
            self.highlighter.add_mapping(pattern,fmt)

        for keyword in ["name", "value"]:
            fmt = QTextCharFormat()
            fmt.setFontWeight(QFont.Bold)
            fmt.setForeground(QtGui.QColor("#ff4f4f"))
            pattern = r'%s\:' % keyword
            self.highlighter.add_mapping(pattern,fmt)


        self.highlighter.setDocument(self.textEdit.document())




if __name__ == "__main__":
    """
    """

    app = QtWidgets.QApplication(sys.argv)
    # window = ProjectDesigner()
    window = SettingsEditor()
    app.exec_()




from sfctools.automation.runner import ModelRunner
from sfctools import Agent,World


from sfctools.examples.example_wrapper import Example



def run():
        
    class SomeAgent(Agent):
        def __init__(self, a):
            super().__init__()
            self.some_attribute = a
    
    my_agent = SomeAgent(a='Hello')
    your_agent = SomeAgent(a='World')
    
    my_agents = World().get_agents_of_type("SomeAgent")
    my_message = my_agents[0].some_attribute
    your_message = my_agents[1].some_attribute

    print("%s says %s, %s says %s" % (my_agent, my_message, your_agent, your_message))



class ReadMeExample(Example):
    def __init__(self):
        super().__init__(lambda: run)
    


if __name__ == "__main__":
    ReadMeExample().run()


__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'dev' # options are: dev, test, prod

import pandas as pd
import numpy as np
import warnings

"""
This file embraces the following features

IndicatorReport
DistributionReport
ReportingSheet

WARNING BETA, under development, use at own risk!
"""

class IndicatorReport:
    """
    A generic class for logging scalar values ('indicators').
    """

    _instances = {}

    def __init__(self,xlabel,ylabel,data=None):
        """
        :param xlabel: x axis label, str
        :param ylabel: y axis label, str
        :param data: some initial data (default None). Has to have 'append' method
        """
        if data is not None:
            self.data = data
        else:
            self.data = []
        self.xlabel = xlabel
        self.ylabel = ylabel

        if not self.ylabel in self.__class__._instances:
            self.__class__._instances[self.ylabel] = self
        else:
            warnings.warn("Tried to add a report with a variable that is already registered. Skipping...")

    @classmethod
    def getitem(cls,key):
        """
        retrieves a certain report from the instances created.

        :param key: the ylabel of the respective report
        """
        return cls._instances[key]

    def add_data(self,x):
        """
        inserts some data into the data structure

        :param x: a scalar value
        """
        self.data.append(x)

    def plot_data(self,ax,s=1.8,color="black",scatter_color="gray"):
        """
        plots the data onto a matplotlib axis

        :param ax: a matplotlib axis
        """

        data = self.data
        x = np.arange(len(data)) + 1
        y = data # np.random.rand(len(x))

        if len(data) > 1 and data[1] is not None:
            has_labels = True

        ax.plot(x, y, "-", color=color) # <- manually optimized

        if len(y) < 20:
            ax.scatter(x, y, color=scatter_color, s=s)  # <- manually optimized

        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ylabel)

        # ax.set_title(self.title)


class DistributionReport:
    """
    A generic class for a logger of a distribution (i.e. )
    """

    def __init__(self,xlabel,ylabel,data=None):
        """
        :param xlabel: x axis title of the report
        :param ylabel: y axis title of the report
        :param data: initial data (default None). Has to have 'append' method, ideally a list of tuples (data_i, tag)
        """
        if data is not None:
            sorted_data = list(data.copy())
            sorted_data.sort(key = lambda x: x[0])
            self.data = sorted_data
        else:
            self.data = []

        self.xlabel = xlabel
        self.ylabel = ylabel

    def add_data(self, x, label=None):
        """
        adds data into the data structure. (stores a sorted version of x).

        :param x: list or numpy array to store
        :param label: some additional tag
        """

        y = np.sort(x.copy())

        # store distribution
        self.data.append((y,label))

    def plot_data(self, ax,color=None,s = None):
        """
        plot the data onto a matplotlib axis

        :param ax: a matplotlib axis
        :param color: plotting color
        :param s: scatter size
        """

        if color is None:
            col = "gray" # colors = ["gray", "indianred", "mediumblue"]
        else:
            col = color

        sorted_data = list(self.data.copy())
        sorted_data.sort(key = lambda x: x[0])

        if s is None:
            s = max(2.0, 8.0 - 4.0 * (len(sorted_data)/1000.0))

        for i,data in enumerate(sorted_data):
            x = data[0]
            y = data[1] # np.random.rand(len(x))

            # #ax.plot(x, y,color=colors[i%len(colors)], alpha=1-0.8*i/len(self.data[0]),label=label)
            ax.scatter(x, y, color=col, s= s)

        ax.set_ylabel(self.xlabel)
        ax.set_xlabel(self.ylabel)
        # ax.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)


class ReportingSheet:
    """
    Reporting sheet is an overview sheet for reporting model results in form of a grid plot.
    """

    def __init__(self,instances=None):
        """
        :param instances: default None, if not None: list of Report instances can be passed here
        """
        if instances is None:
            self.items = []
        else:
            self.items = instances

    def add_report(self,report):
        """
        Adds a report item to this sheet.

        :param report: any IndicatorReport or DistributionReport
        """
        self.items.append(report)

    def plot(self, show_plot=True,verbose=False):
        """
        Generates a plot grid using matplotlib

        :param show_plot: show the figure in a window? default True. If False, figure is returned
        :param verbose: print number of rows and columns if True (default False)
        :return fig: matplotlib figure
        """

        import matplotlib.pyplot as plt
        plt.rcParams.update({'font.size': 8})

        # layout is 2 plots per column...
        
        if len(self.items) == 1:
            n_cols = 1
            n_rows = 1

        else:
            n_cols = 2
            n_rows = max(1,int(np.ceil(len(self.items)/n_cols)))

        if verbose: print("<Reporting Sheet> n_rows","n_cols",n_rows,n_cols)

        cm = 1 / 2.54  # centimeters in inches
        fig = plt.figure(figsize=(0.5*3.1415*10.5*cm,14.8*cm)) # DIN A6

        try:
            plt.style.use("macro")
        except:
            pass

        for i,item in enumerate(self.items):
            ax = fig.add_subplot(n_rows,n_cols,i+1)
            item.plot_data(ax)

        plt.tight_layout()
        fig.subplots_adjust(hspace=.47)

        if show_plot:
            plt.show()

        return fig

    def to_dataframe(self):
        """
        combines the data into a single pandas dataframe
        """
        df = pd.DataFrame()
        for i, item in enumerate(self.items):
            # TODO warn if something is overwritten

            if isinstance(item, IndicatorReport):
                df[item.xlabel] = list(range(len(item.data)))
                df[item.ylabel] = item.data

            elif isinstance(item, DistributionReport):
                df[item.xlabel] = [d[0] for d in item.data]
                df[item.ylabel] = [d[1] for d in item.data]

            else:
                warnigns.warn("Found strang item in Reporting sheet. Skipping. %s" % str(item))
        # df.index.name = self.items[0].xlabel
        return df

    def to_latex(self):
        """
        Generates latex code NOT YET IMPLEMENTED
        """
        raise NotImplementedError("Not yet implemented. Wait for beta release.")


if __name__ == "__main__":

    rep_sheet = ReportingSheet()

    for i in range(3):
        my_dist = DistributionReport(title="Distribution 1")
        my_dist.add_data(np.random.rand(10),label="t=1")
        my_dist.add_data(np.random.rand(10),label="t=2")
        my_dist.add_data(np.random.rand(10), label="t=3")
        rep_sheet.add_report(my_dist)

    for i in range(3):
        my_ind = IndicatorReport(xlabel="Time", ylabel="Some Indicator")
        my_ind.add_data(np.random.rand(10), label="t=1")
        my_ind.add_data(np.random.rand(10), label="t=2")
        my_ind.add_data(np.random.rand(10), label="t=3")
        rep_sheet.add_report(my_ind)

    rep_sheet.plot()




__version__ = "0.9"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'prod' # options are: dev, test, prod


from ..core.settings import Settings
from ..bottomup.stock_manager import StockManager
from ..core.flow_matrix import FlowMatrix

import pandas as pd
import warnings
import numpy as np
from collections import defaultdict
from enum import Enum
import copy


class BalanceEntry(Enum):
    """
    these are the different entry types
    """
    ASSETS = 0
    EQUITY = 1
    LIABILITIES = 2
    TOTAL = 3


class BankruptcyEvent(Enum):
    """
    bankruptcy cases
    """
    NEGATIVE_ASSETS = 0
    NEGATIVE_LIABILITIES = 1
    NEGATIVE_EQUITY = 2


class BalanceSheet:
    """
    This is the balance sheet of an agent (it could be used as a standalone datastructure, too, in theory). A balance sheet is of the following form

+------------------+-------------------------+
|    ASSETS        | LIABILITIES AND EQUITY  |
+==================+=========================+
|                  |                         |
|   ...            |   Liabilities           |
+------------------+-------------------------+
|                  |   ...                   |
+------------------+-------------------------+
|                  |   Equity                |
+------------------+-------------------------+
|                  |   ...                   |
+------------------+-------------------------+
|TOT ASSETS     = TOT LIABILITIES AND EQUITY |
+--------------------------------------------+

    more information can be found here https://www.investopedia.com/terms/b/balancesheet.asp for example.

    """

    _tolerance = 1e-6 # tolerance for engage()
    _tolerance_rel = 1e-4 # relative tolerance of deviation from total assets

    @classmethod
    def change_tolerance(cls, tol:float, rel_tol:float):
        """
        Change the tolerance by which the balace is allowed to diverge
        This will set the new values globally. The tolerance is checked in a two-step process:
        Frist, the absolute tolerance violation is checked, i.e. if any balance sheet side sums up to more or less than the other side.
        If the absolute tolerance is violated, an error is thrown if also the relative tolerance is exceeded. The latter is the second step of the checking process.
        If the relative deviation is violated but the absolute level is not exceeded, no error wil be thrown.

        :param tol: new absolute tolerance
        :param rel_tol: relative tolerance (relative devaition from total assets)
        """

        assert tol > 0
        cls._tolerance = tol
        cls._tolerance_rel = rel_tol

    class Proxy:
        """
        Proxy class for balance sheet. It can be either engaged (consistent and 'active') or disengaged (inconsistent and 'disabled').
        The consistency is automatically checked inside the 'engage()' call when switched from disengaged to engaged.
        """
        def __init__(self, parent):
            self.parent = parent

        def __enter__(self):
            self.parent._engaged = False # disengage()

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.parent.engage()
            self.parent._engaged = True # engage()

    def __init__(self, owner):
        """
        Balance sheet constructor. Creates a new balance sheet

        :param owner: the owner agent of this balance sheet. Can be None for standalone purposes, but it is recommended to pass an Agent instance here.
        """
        self.owner = owner # stores owner instance, if any
        default_value = 0.0 # np.nan NOTE if nan is set this can lead to misbehavior in engage.
        self._sheet = defaultdict(lambda: {BalanceEntry.ASSETS: default_value,
                                          BalanceEntry.EQUITY: default_value,
                                          BalanceEntry.LIABILITIES: default_value}) # the actual data structure is a three-column table for assets, equity and liabilities

        self._sheet[BalanceEntry.TOTAL] = {BalanceEntry.ASSETS: default_value,
                               BalanceEntry.EQUITY: default_value,
                               BalanceEntry.LIABILITIES: default_value} # the 'total' row at the bottom

        self._engaged = True  # <- this can deactivate the consistency check temporarily for modification. Manual modification is not allowed (therefore private).
        self._bankrupt = False # <- this stores if the owner went bankrupt. manual modification should not be needed (therefore private).
        # self.tolerance -> self.__class__.tolerance # numerical tolerance for consistency cross-check

        # link agents to specific entries in the balance sheet
        # nested by [BalanceEntry]->[ItemName]->[Agent]->[float Balance]
        self._changelog = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: 0.0)))

    def add_changelog(self, agent, name, which, x):
        """
        inserts a link between a specific balance sheet item to a
        for example, a bank might want to record which agent its deposits belong to...
        Data structure: [BalanceEntry]->[ItemName]->[Agent]->[float Balance]

        :param agent: agent instance to store here
        :param name: name of the item, e.g. 'Deposits'
        :param which: corresponding BalanceEntry
        :param x: quantity which is inserted, can be positive, zero or negative
        """
        # assert isintance(x,int) or isinstance(x,float), "wrong data type"
        y = self._changelog[which][name][agent]

        if np.isnan(y):
            self._changelog[which][name][agent] = x  # first time this is called? overwrite nan
        else:
            self._changelog[which][name][agent] = x + y # after first time

    @property
    def raw_data(self):
        """
        get the raw data of this data structure in dictionary format.
        This will create a copy of the original data, so the user won't operate on the data directly (which is forbidden)
        """
        d = copy.deepcopy(self._sheet) # deep copy because its a nested dict TODO compress the data somehow
        del d[BalanceEntry.TOTAL]
        return d

    def get_last_changelog(self, name, which,agent=None):
        """
        returns the internally stored links to other agents made via balance sheet transactions

        :param agent: agent instance to store here. if None, all agents' entries will be returned
        :param name: name of the item, e.g. 'Deposits'
        :param which: corresponding BalanceEntry
        :return: float, current balance or dict of balances for agents
        """

        if agent is not None:
            return self._changelog[which][name][agent]
        return self._changelog[which][name]

    #def items(self):
    #    return self.raw_data.items()

    #def values(self):
    #    return self.raw_data.values()

    #def keys(self):
    #    return self.raw_data.keys()

    def __repr__(self):
        return "<Balance Sheet of %s>" % self.owner

    def __str__(self):
        return self.__repr__()

    def __getitem__(self, key):
        """
        Get an entry of the balance sheet
        :param key: str, name of the asset
        :return: a dict with ASSETS, EQUITY, LIABILITIES entries of this item in this particular agent. NOTE In most cases, at least 2 of the entries will be zero.
        """
        return self._sheet[key]

    @property
    def modify(self):
        """Modification decorator. Temporarily disengages the data structure. This can be used in a with block:

        .. code-block:: python

            with balance_sheet_1.modify:
                with balance_sheet_2.modify:
                    balance_sheet_1.change_item(...)
                    balance_sheet_2.change_item(...)

        """
        return self.Proxy(self)

    def restore_after_bankrupcy(self,verbose=False):
        # helper function to reset bankrupt flag to False again after bankruptcy has been fixed (e.g. replacement of agent by new fictitious entrant)
        if verbose:
            warnings.warn("%s restore after bankrupcy" % self.to_string()) # TODO < test this

        self._bankrupt = False

    def disengage(self):
        """
        'unlocks' the balance sheet -> can be modified now without error
        """
        self._engaged = False

    def engage(self):
        """
        locks the balance sheet -> no longer can be modified
        performs also a cross-check if consistency is maintained
        """

        eps = self.__class__._tolerance # tolerance level for cross-check, 1e-6 by default

        a = 0.0 # dummy for sum of entries
        e = 0.0
        l = 0.0
        
        # iterate through the rows of the balance sheet and sum up assets, liabilities, equity
        for entry, data in self._sheet.items():

            if entry == BalanceEntry.TOTAL:
                continue

            a_i = data[BalanceEntry.ASSETS]
            e_i = data[BalanceEntry.EQUITY]
            l_i = data[BalanceEntry.LIABILITIES]

            a += a_i
            e += e_i
            l += l_i

            cond_a = isinstance(a,int) or isinstance(a,float)
            cond_e = isinstance(e,int) or isinstance(e,float)
            cond_l = isinstance(l,int) or isinstance(l,float)

            type_cond = cond_a and cond_e and cond_l

            if type_cond and a_i < -eps: # handle negative assets

                if True: #  TODO might add exceptions for some special cases?
                    if not self._bankrupt:
                        self._bankrupt = True
                        warnings.warn("filed bankruptcy for %s\n" %self.owner)
                        self.owner.file_bankruptcy(event=BankruptcyEvent.NEGATIVE_ASSETS)

                    return None

                else: # NOTE not used as of now...
                    s = self.to_string()
                    error = ValueError("%s of %s reached a value below zero after engage() call" % (entry, self.owner) + "\n\n" + s)
                    raise error

            if type_cond and l_i < -eps: # handle negative liabilities

                if True: #  TODO might add exceptions for some special cases?
                    if not self._bankrupt:
                        self._bankrupt = True
                        warnings.warn("filed bankruptcy for %s\n" %self.owner)
                        self.owner.file_bankruptcy(event=BankruptcyEvent.NEGATIVE_LIABILITIES)

                    return None

                else: # NOTE not used as of now...
                    s = self.to_string()
                    error = ValueError("%s of %s reached a value below zero after engage() call" % (entry, self.owner) + "\n\n" + s)
                    raise error

        cond_a = isinstance(a,int) or isinstance(a,float)
        cond_e = isinstance(e,int) or isinstance(e,float)
        cond_l = isinstance(l,int) or isinstance(l,float)
        type_cond = cond_a and cond_e and cond_l

        if type_cond and  abs(a - (e+l)) > eps: # check bookkeeping equation for consistency within tolerance level eps

            # almost zero? bookkeeping equation obeyed
            eps2 = self.__class__._tolerance_rel # relative tolerance
            if a !=0 and abs(a - (e+l))/abs(a) < eps2:

                pass
                # relative deviation is still small?  -> just warn
                # warnings.warn("balance sheet of %s is numerically instable. Deviation %.10f" % (self.owner, a-(e+l)) + "\n\n" + s)
                # NOTE ^ above code was commented out because no case could be thought of where this would make sense.

            else: # not almost zero? bookkeping equation violated
                s = self.to_string()
                error = ValueError("balance sheet of %s is corrupted after cross-check. Deviation %.10f" % (self.owner, a-(e+l)) + "\n\n" + s)
                raise error

        if type_cond and e < 0 and e < -eps: # negative equity? = 'balance sheet under water'

            if not self._bankrupt:
                self._bankrupt = True
                self.owner.file_bankruptcy(event=BankruptcyEvent.NEGATIVE_EQUITY)
            else:
                # already bankrupt -> do not trigger bankruptcy a second time... # TODO < think about logic in here
                pass
            return None

    def get_balance(self, key, kind=BalanceEntry.ASSETS) -> float:
        """
        Gets amount of a certain entry for certain key

        :param key: name of entry (e.g. 'Cash' or 'Apples')
        :param kind: preferrably BalanceEntry: Asset, Liabilities or Equity, optionally str 'Assets', 'Equity' or 'Liabilities'

        :return: float, value in balance sheet
        """

        # TODO automatically detect which kind it is if the other two are zero(?) instead of default value being ASSETS

        if isinstance(kind, str):
            mydict = {
                "Assets": BalanceEntry.ASSETS,
                # BalanceEntry.ASSETS: BalanceEntry.ASSETS,
                "Equity": BalanceEntry.EQUITY,
                # BalanceEntry.EQUITY: BalanceEntry.EQUITY,
                "Liabilities": BalanceEntry.LIABILITIES,
                # BalanceEntry.LIABILITIES: BalanceEntry.LIABILITIES,
            }
            which = mydict[kind]
        else:
            which = kind # should be BalanceEntry at this point! Will raise KeyError if not

        if not which in self._sheet[key]:
            raise KeyError("Could not find entry for %s"% which)

        return_value = self._sheet[key][which]

        if np.isnan(return_value): # nan values are not allowed, so throw an error
            msg = "Requested Value is NaN (not a number)."
            raise RuntimeError(msg)

        return return_value

    def to_string(self, nice_format=True):
        """
        Coverts the whole balance sheet to a pandas dataframe and subsequently to a string

        :param nice_format: True (default), use a fancy formatting in three boxes vs direct dataframe to string conversion
        :return: str, balance sheet string representation
        """
        if nice_format:

            def filter_empty(series):
                #try:
                indices = [str(v).strip() != ".-" for v in series]
                return series[indices]
                #except:
                #    return df 
            
            # 1. retrieve strings from all 3 components
            df = self.to_dataframe()
            df.index.name = ""
            df_a = "ASSETS\n-------\n" + filter_empty(df["Assets"]).to_string().replace("NaN"," ~ ")
            df_l = "LIABILITIES\n-------------\n" + filter_empty(df["Liabilities"]).to_string().replace("NaN"," ~ ")
            df_e = "EQUITY/NET WORTH\n------------------\n" + filter_empty(df["Equity / Net Worth"]).to_string().replace("NaN"," ~ ")

            # 2. 'glue' together the string blocks 
            df_el = df_l + "\n\n" + df_e

            lines_a = df_a.split("\n")
            lines_l = df_el.split("\n")

            max_len_a = np.max([len(i) for i in lines_a])
            max_len_l = np.max([len(i) for i in lines_l])
            
            final_str = "-"*(max_len_a + max_len_l + 5)
            final_str += "\n%s" % self + "\n\n"
            # final_str += "ASSETS" + " "*(max_len_a-2) +  "LIABILITIES\n"
            # final_str += "------" + " "*(max_len_a-2) +  "-----------\n"

            title_switch = False
            for i in range(max(len(lines_a), len(lines_l))):
                if i < len(lines_a):
                    final_str += lines_a[i] + " "*(max_len_a-len(lines_a[i]))
                
                else:
                    final_str += " "*(max_len_a)

                final_str += "  |  "
                
                if i < len(lines_l):
                    final_str += lines_l[i]
                
                final_str += "\n"

            final_str += "-"*(max_len_a + max_len_l + 5) + "\n"
            return final_str 



        else:
            return "\n\n" + self.to_dataframe().to_string().replace("NaN"," ~ ") + "\n\n"


    def change_item(self, name, which, value, suppress_stock=False):
        """
        Chanes value of an item in the balance sheet.

        :param name: name of the asset / entry to change
        :param value: value is added to entry. can also be negative
        :param which: BalanceEntry: Asset, Liability, Equity
        """

        type_cond = isinstance(value,int) or isinstance(value,float)

        if type_cond:
            if value == 0:  # security lock if this is a 'ghost transaction'
                warnings.warn("Blocked ghost item in balance (value=0): %s, %s,%s,%s"% (self,name,which,value))
                return

            if self._engaged: # security lock when balance is still disengaged
                raise PermissionError("Cannot change item in engaged balance sheet! This would lead to inconsistencies. Please disengage before calling change_item or use 'modify' proxy.")

            if np.isnan(self._sheet[name][which]):
                warnings.warn("Found NaN value in %s, %s, %s" % (self,name,which))
                # self._sheet[name][which] = 0 # NOTE <- alternative way to handle

            if np.isnan(self[BalanceEntry.TOTAL][which]):
                warnings.warn("Found NaN value in %s, %s, %s" % (self,"TOTAL",which))
                # self._sheet[BalanceEntry.TOTAL][which] = 0 # NOTE <- alternative way to handle

        self._sheet[name][which] += value
        self._sheet[BalanceEntry.TOTAL][which] += value
        
        if not suppress_stock:
            # Suppresses logging in the transaction flow matrix. This can be helpful in some cases

            if which == BalanceEntry.ASSETS:
                FlowMatrix().capital_flow_data["Δ %s" % name][self.owner] += -value #-value
            if which == BalanceEntry.LIABILITIES:
                FlowMatrix().capital_flow_data["Δ %s" % name][self.owner] += value #value

            if which == BalanceEntry.EQUITY and name == "Cash":
                raise ValueError("Cash cannot be equity")


    """
    TODO (?) insert operations like 'Aktivtausch', 'Passivtausch', 'Bilanzverlängerung'(?)
    """

    def to_dataframe(self):
        """
        Create a dataframe from this data structure. Warning: this is computationally heavy and should not be used in loops!

        :return: pandas dataframe
        """

        df = pd.DataFrame({"Entry": [],"Assets": [],"Liabilities": [], "Equity / Net Worth": []}) # construct an empty 'dummy' data frame
        entries = [] # allocate list of commodities entered in the balance sheet

        null_sym = "   .-   " # symbol for zero entries

        # iterate through the balance sheet items and 'fill the table'
        for key, row in self._sheet.items():

            if key != BalanceEntry.TOTAL: # consider all rows except the 'total' entry
                entries.append(key) # add to entries list
                new_row = pd.DataFrame({"Entry": [key],
                                        "Assets": [row[BalanceEntry.ASSETS]],
                                        "Equity / Net Worth": [row[BalanceEntry.EQUITY]],
                                        "Liabilities": [row[BalanceEntry.LIABILITIES]]}) # construct a new row
                df =  pd.concat([df,new_row.replace(0.0,null_sym)])

        df = df.sort_index() # sort the entries by index

        # Compute 'Total' row
        # dummy for modifying factors.
        # NOTE is 1 always for now, but might change with exchange rates or any other conversion factors (?)
        factors = [1.0 for key in entries]
        my_row = pd.DataFrame({"Entry": ["Total"],
                                "Assets": [np.dot(df["Assets"].replace(null_sym,0.0),factors)],
                                "Equity / Net Worth": [np.dot(df["Equity / Net Worth"].replace(null_sym,0.0),factors)],
                                "Liabilities": [np.dot(df["Liabilities"].replace(null_sym,0.0),factors)]})
        df = pd.concat([df,my_row]) # adds the 'Total' row

        df = df.set_index("Entry") # sets index column correctly
        df.index.name = "BALANCE SHEET OF %s" % self.owner # gives the balance sheet an understandable title

        return df

    def depreciate(self):
        """
        (BETA) this will look up all the required depreciation rates in the
        settings and depreciate the balance sheet.
        Equity will become less, liabilities will stay.
        """

        settings = Settings() # read depreciation rates from settings

        for item_name in self._sheet.keys():

            depr_rate = settings.config_data["params"][item_name]["depreciation"]
            # Settings entry is 'param->(item name)->depreciation->value'

            depr_q = self._sheet[item_name][BalanceEntry.EQUITY] * depr_rate
            # Warning: Depreciation is percentage of current stock. Non-linear!

            if not np.isnan(self._sheet[item_name][BalanceEntry.EQUITY]):
                self._sheet[item_name][BalanceEntry.EQUITY] -= depr_q

            if not np.isnan(self._sheet[item_name][BalanceEntry.ASSETS]):
                self._sheet[item_name][BalanceEntry.ASSETS] -= depr_q

        # TODO maybe automatically log this on income statement of the underlying agent here?

    @property
    def leverage(self) -> float:
        """
        computes the financial leverage value (debt-to-assets ratio)
        of this balance sheet

        .. math::
            Leverage = LIABILITIES / (LIABILITIES + NET WORTH)

        :return: float, leverage
        """

        E = self.net_worth
        L = self.total_liabilities

        if E + L == 0:
            return np.inf

        lev = L / (E+L)

        return max(0,lev)

    @property
    def net_worth(self) -> float:
        """net worth = sum of equity
        """
        return_val =  self._sheet[BalanceEntry.TOTAL][BalanceEntry.EQUITY] #"Equity"]
        if np.isnan(return_val):
            return 0.0
        return return_val

    @property
    def total_assets(self):
        """
        total assets = sum of assets column
        """
        return_val = self._sheet[BalanceEntry.TOTAL][BalanceEntry.ASSETS] # "Assets"]
        if np.isnan(return_val):
            return 0.0
        return return_val

    @property
    def total_liabilities(self):
        """
        total liabilities = sum of all liabilities in liability column
        """
        return_val = self._sheet[BalanceEntry.TOTAL][BalanceEntry.LIABILITIES] # "Liabilities"]
        if np.isnan(return_val):
            return 0.0
        return return_val


    def plot(self,show_labels=True,cols_assets=None, cols_liabs=None,cols_eq=None,show_legend=True,
                  fname=None,ax=None,label_fmt="{0:.1f}"):
        """
        creates a matplotlib plot of the balance sheet

        :param show_labels: show labels (numbers) above the bars
        :param label_fmt: (default '{0:.1f}'), format of labels (if shown)
        :param show_legend: (True) show a legend with the balance sheet entries

        By default, assets are red, equity is blue and liabilites are green
        but custom colors can be given

        :param cols_assets: (None) list of mpl colors for assets. Default is ["salmon", "indianred", "rosybrown", "orangered", "red", "indianred"]
        :param cols_liabs: (None) list of mpl colors for liabilities. Default is [ "lightgreen", "honeydew","seagreen", "darkgreen", "darkslategray"]
        :param cols_eq: (None) list of mpl colors for equity. Default is  ["lightsteelblue", "lavender", "powderblue", "steelblue", "royalblue", "blue"]

        if you want to save to file instead of showing as plot window, specify a fname
        :param fname: if None (default) is given, plot will be shown directly. If fname is given, plot will be saved to a file


        if you want to redirect the plot to a certain axis, you can enter the axis via the parameter ax
        :param ax: axis to plot on. setting this parameter different from None will make 'fname' ineffective
        """
        import matplotlib.pyplot as plt
        import matplotlib.patches as mpatches
        from ..misc.mpl_plotting import matplotlib_barplot

        balance_sheet = self

        title = "%s" % balance_sheet.owner

        assets = []
        liabilities = []

        colors = []

        if cols_liabs is None:
            greens = [ "lightgreen", "honeydew","seagreen", "darkgreen", "darkslategray"]
        else:
            greens = cols_liabs

        if cols_eq is None:
            blues = ["lightsteelblue", "lavender", "powderblue", "steelblue", "royalblue", "blue"]
        else:
            blues = cols_eq

        if cols_assets is None:
            reds = ["salmon", "indianred", "rosybrown", "orangered", "red", "indianred"]
        else:
            reds = cols_assets

        greens.reverse()
        blues.reverse()
        reds.reverse()

        # color patches for legend
        legend_patches = []
        blue_patches = []
        green_patches = []
        red_patches = []

        # data for plotting
        data = {"Assets": [], "Liabilities & Equity": []}
        labels = {}
        i = 0

        assets = []
        lia = []
        equi = []

        # construct data for plotting
        for key, val in balance_sheet._sheet.items():

            if key != BalanceEntry.TOTAL:

                if val[BalanceEntry.LIABILITIES] != 0.0:  # is liability

                    # data["Assets"].append(0.0)
                    # data["Liabilities & Equity"].append(val["Liabilities"])
                    assets.append([key, 0.0])
                    lia.append([key, val[BalanceEntry.LIABILITIES]])

                if val[BalanceEntry.EQUITY] != 0.0:  # is Equity

                    # data["Assets"].append(0.0)
                    # data["Liabilities & Equity"].append(val["Equity"])

                    assets.append([key, 0.0])
                    equi.append([key, val[BalanceEntry.EQUITY]])

                if val[BalanceEntry.ASSETS] != 0.0:  # is Asset

                    # data["Assets"].append(val["Assets"])
                    # data["Liabilities & Equity"].append(0.0)

                    assets.append([key, val[BalanceEntry.ASSETS]])
                    lia.append([key, 0.0])

        # construct data for plotting
        # by default, assets are red, equity is blue and liabilites are green
        k = 0
        for a in assets:

            key = a[0]
            val = a[1]

            if val != 0:
                data["Assets"].append(val)
                data["Liabilities & Equity"].append(0.0)

                col = reds[k%len(reds)]
                colors.append(col) # (col, "black"))
                red_patches.append(mpatches.Patch(color=col, label=key))

                labels[i] = key
                i += 1

            k+=1


        k = 0
        for e in equi:
            key = e[0]
            val = e[1]

            if val != 0:
                data["Assets"].append(0.0)
                data["Liabilities & Equity"].append(val)

                col = blues[k%len(blues)]
                colors.append(col) # ("black", col))
                blue_patches.append(mpatches.Patch(color=col, label=key))

                labels[i] = key
                i += 1

            k+=1

        k = 0
        for l in lia:

            key = l[0]
            val = l[1]

            if val != 0:
                data["Assets"].append(0.0)
                data["Liabilities & Equity"].append(val)

                col = greens[k%len(greens)]
                colors.append(col) #("black", col))
                green_patches.append(mpatches.Patch(color=col, label=key))

                labels[i] = key
                i += 1

            k+=1

        # create customized patche for the legend
        legend_patches = red_patches + blue_patches + green_patches

        df = pd.DataFrame(data).T
        df = df.rename(columns=labels).T

        # print("PLOT\n",df)
        # print("colors",colors)



        fig = matplotlib_barplot(df.T, xlabel="", ylabel="", color=colors, title=title, stacked=True, show_labels=show_labels,fmt=label_fmt,
                                legend="off", size=(4, 4), show=False,ax=ax)

        if ax is None:
            ax = plt.gca()

        legend_patches.reverse()
        ax.legend(handles=legend_patches, bbox_to_anchor=(1.3, 1),framealpha=0.0)

        plt.axis("off")
        plt.tight_layout()

        if ax is None:
            if fname is None:
                plt.show()
            else:
                plt.savefig(fname)
                plt.close()

        return ax

    @classmethod
    def plot_list(cls,list_of_balance_sheets,dt=1,xlabel=None,ylabel=None,title=None,show_liabilities=True,show=True):
        """
        Plots a list of balance sheets (e.g. a collected temporal series)

        :param list_of_balance_sheets: a list of BalanceSheet instances
        :param dt: step, (how many values to skip in between)
        :param xlabel: x axis label
        :param ylabel: y axis label
        :param title: title of the figure
        :param show_liabilities: boolean switch to show passive side of balance sheet as downward-pointing bars (default True)
        :param show: show the plot in a new window? if False, only the matplotlib figure is returned
        """
        import matplotlib.pyplot as plt
        import matplotlib.patches as mpatches
        from ..misc.mpl_plotting import matplotlib_barplot

        entries = {}

        for data_i in list_of_balance_sheets:

            # print("DATA_I",data_i)
            #print("\n\n --\n\n")

            for k,v in data_i.items():

                if k not in entries:
                    entries[k] = {}

                for k2, v2 in v.items():

                    if k2 not in entries[k]:
                        entries[k][k2] = []

                    if not np.isnan(v2):

                        if k2 == BalanceEntry.ASSETS:
                            if np.abs(v2) > 0.0:
                                entries[k][k2].append(np.abs(v2))
                        else:
                            if show_liabilities:
                                if np.abs(v2) > 0.0:
                                    entries[k][k2].append(-np.abs(v2))


        # print("ENTRIES",entries)


        plt.figure()

        if xlabel is not None:
            plt.xlabel(xlabel)
        if ylabel is not None:
            plt.ylabel(ylabel)
        if title is not None:
            plt.title(title)

        maxy = 0.01 # np.-1inf
        miny = -0.01 # np.inf

        try:
            plt.style.use("macro")
        except:
            pass

        plt.grid(False)
        # plot upper half

        colors = ["red","maroon","lightcoral","gold","crimson","orange"] # plot colors
        color_idx = 0
        last_vals = None

        for k,v in entries.items():

            for k2,v2 in v.items():

                if  k2 == BalanceEntry.ASSETS:
                    if len(v2) > 0:
                        idx = [int(i) for i in np.arange(0,len(v2)-1,dt)]
                        # idx2 = [str(Clock().get_i) for i in idx]
                        idx2 = [str(i) for i in idx]

                        #print("idx2",idx2)
                        vals = [v2[i] for i in idx]
                        # print("vals",vals)

                        maxy = max(maxy,np.max(vals))
                        miny = min(miny,np.min(vals))

                        plt.bar(idx2,vals,label=str(k),color=colors[color_idx],alpha=0.9,bottom=last_vals)

                        if last_vals is None:
                            last_vals = np.array(vals)
                        else:
                            last_vals += np.array(vals)

                        color_idx += 1
                        color_idx = color_idx % len(colors)


        colors = ["black","blue","navy","dodgerblue","royalblue","slateblue"] # plot colors
        color_idx = 0
        last_vals = None

        for k,v in entries.items():
            for k2,v2 in v.items():

                if  not k2 == BalanceEntry.ASSETS:

                    if len(v2) > 0:
                        idx = [int(i) for i in np.arange(0,len(v2)-1,dt)]
                        # idx2 = [str(Clock().get_i) for i in idx]
                        idx2 = [str(i) for i in idx]

                        #print("idx2",idx2)
                        vals = [v2[i] for i in idx]
                        # print("vals",vals)

                        maxy = max(maxy,np.max(vals))
                        miny = min(miny,np.min(vals))

                        plt.bar(idx2,vals,label=str(k),color=colors[color_idx],alpha=0.9,bottom=last_vals)

                        if last_vals is None:
                            last_vals = np.array(vals)
                        else:
                            last_vals += np.array(vals)

                        color_idx += 1
                        color_idx = color_idx % len(colors)

        # downward ticks
        # https://stackoverflow.com/questions/50571287/how-to-create-upside-down-bar-graphs-with-shared-x-axis-with-matplotlib-seabor

        plt.xticks(rotation=90)

        #if show_liabilities:
        #    plt.ylim([1.3*miny,1.8*maxy+ .2*max(0,len(entries.keys())-3)])
        #else:
        #    plt.ylim([1.3*miny,1.4*maxy+ .2*max(0,len(entries.keys())-3)])


        ticks =  plt.gca().get_yticks()
        myticks_abs = [(abs(tick)) for tick in ticks]
        myticks = [((tick)) for tick in ticks]
        plt.gca().set_yticks(myticks)
        plt.gca().set_yticklabels(myticks_abs)

        plt.legend(loc=(1.04,0)) # "upper left")

        plt.tight_layout()
        if show:
            plt.show()

        return plt.gcf()




"""
__author__ Thomas Badauf, Ve
__date__ March 2023

This is a wrapper class to run an example instance.
"""

import time

class Example():
    """
    This is a wrapper class to run an example instance.
    """

    def __init__(self,runner_method):
        """
        :param runner_method: a method to execute in this example
        """
        self.runner_method = runner_method
        self.t_start = None
        self.t_end = None
        self.rtime = None
        self.res = None

    def run(self,verbose=True):
        """
        run the actual example
        """
        self.t_start = time.time()
        self.res = self.runner_method()
        self.t_end = time.time()

        try:
            self.rtime = self.t_end - self.t_start
        except:
            pass

        if verbose:
            self.print_stats()

        return self.res


    def print_stats(self):
        """
        print some statistics,
        """

        if self.t_start is None or self.t_end is None:
            print("< No stats available. > ")
        
        statstr = """
        \n\n
        STATS FOR %s
        -----------------------------
          runtime (s):  %.2f
          result type:  %s
        -----------------------------
        """  % (self.__class__.__name__,self.rtime,type(self.res))

        print(statstr)



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'prod' # options are: dev, test, prod

import os 

def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    """
    This defines the warning format of all sfctools warnings.
    """
    return '%s %s: %s [%s LINE %i]\n' % ("[sfctools] ", category.__name__, message,os.path.basename(filename),lineno)





from .balance import BalanceSheetExample
from .order_book import BankOrderBookExample
from .basic_example.basic_example import BasicExample
from .clock import ClockExample
from .exampleabm.model import ABMExample
from .hello_agent import HelloWorldExample
from .market_example.market import MarketExample
from .market_example.market2 import MarketExample2
from .monte_carlo.monte_carlo import MonteCarloExample
from .signal_slot import SignalSlotExample
from .inventory import InventoryExample
from .flowmatrix import FlowMatrixExample
from .readme_example import ReadMeExample


__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'prod' # options are: dev, test, prod


from collections import defaultdict, deque
from functools import partial
import numpy as np
from enum import Enum

class CollectionType(Enum):
    SINGLE = 0
    LIST = 1
    DEQUE = 2


class Collection:
    """
    A 'collection' is a special data structure: it is a dict of items.

    kind 'single' looks like
    {tag: value,
    tag2: value,
    tag3: value,
    ...}

    inserting a new value will result in overwrite.

    kind 'list' looks like
    {tag: [value,value,value,..],
    tag2: [value,value,...],
    tag3: [...],
    ...}

    if maxlen is not infinity then insertion will result in a deque append, otherwise list append.

    possible operations are:

    - insert()  inserts a new item
    - remove()   removes an item
    - get_count() gets the number of items stored under a tag
    - get_last() gets the last inserted item of a tag

    """

    def __init__(self, kind="single", maxlen=np.inf):
        """
        :param kind: 'single' or 'list'
        :param maxlen: maximum allowed length. only has an effect if kind is 'list'

        constructor for Collection.
        """

        self.last_tag = None
        self.kind = None
        self._data = None

        """
        assign kind of data format, 
        assign actual data initially
        """

        assert kind == "single" or kind == "list"

        if kind == "list" and maxlen == np.inf:  # a list with infinite length
            self.kind = CollectionType.LIST
            self._data = defaultdict(list)

        elif kind == "list" and maxlen < np.inf:  # a 'list with limited length' (deque)
            self.kind = CollectionType.DEQUE
            self._data = defaultdict(partial(deque, maxlen=maxlen))

        elif kind == "single":  # a single entry
            self.kind = CollectionType.SINGLE
            self._data = {}

        else:
            raise TypeError("Invalid kind of collection. Allowed are 'single' or 'list'")

    def items(self):
        return self._data.items()

    def keys(self):
        return self._data.keys()

    def insert(self, t, new_data, verbose=False):
        """
        insert new data at tag t.

        :param t: the tag (e.g. time) for the data
        :param new_data:  the new data
        :param verbose: verbose flag
        """
        if verbose: print("insert", t, new_data,self.kind)

        if self.kind == CollectionType.SINGLE:
            # if t in self._data:
            #    raise KeyError("Key %s already written" % t)
            # else:
            self._data[t] = new_data

        elif self.kind == CollectionType.LIST or self.kind == CollectionType.DEQUE:
            self._data[t].append(new_data)  # defaultdict will take care of missing keys

        self.last_tag = t

    def remove(self, t, entry):
        """
        remove an entry from a tag t
        :param t:  tag
        :param entry:  entry to be removed
        """

        if t in self._data:
            # self._data.pop(t, None) <- this actually removes all entries
            self._data[t].remove(entry)

            if len(self._data[t]) == 0: # if this was the last item <- pop out the tag completely
                self._data.pop(t,None)

        else:
            raise KeyError("Cannot remove item: Tag %s was not found" % t)  # we do not want to use defaultdict here

    def get_count(self, t):
        """
        Get the number of existing entries for this count
        :param t:
        :return:
        """

        count = 0
        if t in self._data:
            count = len(self._data[t])

        return count

    def get_last(self):
        """
        get the last tag and data. If there was None, return None

        :return: t, data or None
        """

        if self.last_tag is None:
            return None
        else:
            return self.last_tag, self._data[self.last_tag]

    def values(self):
        """
        like in a dict, return all value entries
        """

        all_vals = []

        for key,val in self._data.items():
            if isinstance(val,dict) or isinstance(val,defaultdict):
                all_vals.append(val.values())
            else:
                all_vals.append(val)

        return all_vals

    def get_entries(self):
        return self._data

    def __getitem__(self,key):
        """
        element search in Collection by key

        :param key: key to search for
        :returns: value, stored value
        """

        return self._data[key]


    def __str__(self):
        return str(self._data)